/**
 * 
 */
package org.biojavax.bio.seq.io;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.StringReader;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;
import java.util.regex.Pattern;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.biojava.bio.BioException;
import org.biojava.bio.seq.Feature;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.io.ParseException;
import org.biojava.bio.seq.io.SeqIOListener;
import org.biojava.bio.seq.io.SymbolTokenization;
import org.biojava.bio.symbol.IllegalSymbolException;
import org.biojava.bio.symbol.SimpleSymbolList;
import org.biojava.bio.symbol.Symbol;
import org.biojava.bio.symbol.SymbolList;
import org.biojava.utils.ChangeVetoException;
import org.biojava.utils.xml.PrettyXMLWriter;
import org.biojava.utils.xml.XMLWriter;
import org.biojavax.CrossRef;
import org.biojavax.Namespace;
import org.biojavax.Note;
import org.biojavax.RankedCrossRef;
import org.biojavax.RichAnnotation;
import org.biojavax.RichObjectFactory;
import org.biojavax.SimpleCrossRef;
import org.biojavax.SimpleNote;
import org.biojavax.SimpleRankedCrossRef;
import org.biojavax.SimpleRichAnnotation;
import org.biojavax.bio.seq.RichFeature;
import org.biojavax.bio.seq.RichSequence;
import org.biojavax.bio.seq.SimplePosition;
import org.biojavax.bio.seq.SimpleRichLocation;
import org.biojavax.bio.seq.io.RichSequenceFormat.BasicFormat;
import org.biojavax.bio.taxa.NCBITaxon;
import org.biojavax.bio.taxa.SimpleNCBITaxon;
import org.biojavax.ontology.ComparableTerm;
import org.biojavax.utils.StringTools;
import org.biojavax.utils.XMLTools;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * Format reader for SeqXML files. This version of SeqXML format will generate
 * and write RichSequence objects. SeqXML is a lightweight XML alternative to FASTA.
 * For more details see http://seqXML.org
 * 
 * 
 * @author Sanjit Roopra, Thomas Schmitt
 * 
 */
public class SeqXMLFormat extends BasicFormat {


	// Register this format with the format auto-guesser.
	static {
		RichSequence.IOTools.registerFormat(SeqXMLFormat.class);
	}

	/**
	 * The name of this format
	 */
	public static final String SEQXML_FORMAT = "SeqXML";

	protected static final String SEQXML_TAG = "seqXML";
	protected static final String SEQXML_VERSION ="0.3";
	protected static final String XSI_SCHEMA_LOCATION_ATTR = "xsi:noNamespaceSchemaLocation";
	protected static final String SEQXML_SCHEMA ="http://www.seqXML.org/%s/seqxml.xsd";
	protected static final String SEQXML_SOURCE_ATTR = "source";
	protected static final String SEQXML_SOURCEVERSION_ATTR = "sourceVersion";
	protected static final String SEQXML_SEQXMLVERSION_ATTR = "seqXMLversion";

	protected static final String ENTRY_TAG = "entry";
	protected static final String ENTRY_ID_ATTR = "id";
	private static final String ENTRY_SOURCE_ATTR = "source";

	protected static final String SPECIES_TAG = "species";
	protected static final String SPECIES_NCBITAXID_ATTR = "ncbiTaxID";
	protected static final String SPECIES_NAME_ATTR = "name";
	protected static final String SPECIES_NAMECLASS = "scientific";

	protected static final String DESCRIPTION_TAG = "description";

	protected static final String DBREF_TAG = "DBRef";
	protected static final String DBREF_SEQTYPE_ATTR = "type";
	protected static final String DBREF_SOURCE_ATTR = "source";
	protected static final String DBREF_ID_ATTR = "id";

	protected static final String PROPERTY_TAG = "property";
	protected static final String PROPERTY_NAME_ATTR = "name";
	protected static final String PROPERTY_VALUE_ATTR = "value";

	protected static final String RNASEQ_TAG = "RNAseq";
	protected static final String AASEQ_TAG = "AAseq";
	protected static final String DNASEQ_TAG = "DNAseq";


	
	protected static final Pattern xmlSchemaPattern = Pattern.compile("^<seqXML.*");

	private boolean readHeader = true;
	private String[] headerData = new String[3];

	/**
	 * Implements some UniProtXML-specific terms.
	 */
	public static class Terms extends RichSequence.Terms {
		/**
		 * Getter for the SeqXML term
		 * 
		 * @return The SeqXML Term
		 */
		public static ComparableTerm getSeqXMLTerm() {
			return RichObjectFactory.getDefaultOntology().getOrCreateTerm(
					"SeqXML");
		}

		/**
		 * Getter for the feature value term
		 * 
		 * @return The feature value Term
		 */
		public static ComparableTerm getFeatureValueTerm() {
			return RichObjectFactory.getDefaultOntology().getOrCreateTerm(
					"feature_value");
		}

		/**
		 * Getter for the source name term
		 * 
		 * @return The source name Term
		 */
		public static ComparableTerm getSourceNameTerm() {
			return RichObjectFactory.getDefaultOntology().getOrCreateTerm(
					"source_name");
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean canRead(BufferedInputStream stream) throws IOException {
		stream.mark(2000);
		BufferedReader br = new BufferedReader(new InputStreamReader(stream));
		br.readLine(); // skip first line
		String secondLine = br.readLine();
		boolean readable = secondLine != null
				&& xmlSchemaPattern.matcher(secondLine).matches(); // check on second
		// line
		// don't close the reader as it'll close the stream too.
		stream.reset();
		return readable;
	}

	/**
	 * 
	 */
	public boolean canRead(File file) throws IOException {
		BufferedReader br = new BufferedReader(new FileReader(file));
		br.readLine(); // skip first line
		String secondLine = br.readLine();
		boolean readable = secondLine != null
				&& xmlSchemaPattern.matcher(secondLine).matches(); // check on second
		// line
		br.close();
		return readable;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.biojavax.bio.seq.io.RichSequenceFormat#guessSymbolTokenization(java
	 * .io.BufferedInputStream)
	 */
	@Override
	public SymbolTokenization guessSymbolTokenization(BufferedInputStream arg0)
			throws IOException {
		return RichSequence.IOTools.getProteinParser();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.biojavax.bio.seq.io.RichSequenceFormat#guessSymbolTokenization(java
	 * .io.BufferedInputStream)
	 */
	@Override
	public SymbolTokenization guessSymbolTokenization(File file)
			throws IOException {
		return RichSequence.IOTools.getProteinParser();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.biojava.bio.seq.io.SequenceFormat#readSequence(java.io.BufferedReader
	 * , org.biojava.bio.seq.io.SymbolTokenization,
	 * org.biojava.bio.seq.io.SeqIOListener)
	 */
	@Override
	public boolean readSequence(BufferedReader reader,
			SymbolTokenization symParser, SeqIOListener listener)
			throws BioException, IllegalSymbolException, IOException {
		if (!(listener instanceof RichSeqIOListener))
			throw new IllegalArgumentException(
					"Only accepting RichSeqIOListeners today");
		return this.readRichSequence(reader, symParser,
				(RichSeqIOListener) listener, null);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seeorg.biojavax.bio.seq.io.RichSequenceFormat#readRichSequence(java.io.
	 * BufferedReader, org.biojava.bio.seq.io.SymbolTokenization,
	 * org.biojavax.bio.seq.io.RichSeqIOListener, org.biojavax.Namespace)
	 */
	@Override
	public boolean readRichSequence(BufferedReader reader,
			SymbolTokenization symParser, RichSeqIOListener rlistener,
			Namespace ns) throws BioException, IllegalSymbolException,
			IOException {
		try {
			DefaultHandler header_handler = new SeqXMLHeaderHandler(headerData);
			if (readHeader) {
				readFirstSeqXMLHeader(reader, header_handler);
				readHeader = false;
			}
			DefaultHandler m_handler = new SeqXMLHandler(this, symParser,
					rlistener, ns, headerData);

			return XMLTools.readXMLChunk(reader, m_handler, ENTRY_TAG);
		} catch (ParserConfigurationException e) {
			throw new ParseException(e);
		} catch (SAXException e) {
			throw new ParseException(e);
		}
	}

	private void readFirstSeqXMLHeader(BufferedReader reader,
			DefaultHandler mHandler) throws IOException,
			ParserConfigurationException, SAXException {
		StringBuffer buffer = new StringBuffer();
		String line = null;
		boolean headerLine = false;
		reader.mark(10000);
		while (!headerLine && (line = reader.readLine()) != null) {
			if (xmlSchemaPattern.matcher(line).matches()) {
				buffer.append(line);
				buffer.append("</" + SEQXML_TAG + ">");
				headerLine = true;
			}
		}
		reader.reset();
		SAXParser h_Parser;
		SAXParserFactory factory = SAXParserFactory.newInstance();
		h_Parser = factory.newSAXParser();
		InputSource source = new InputSource(
				new StringReader(buffer.toString()));
		h_Parser.parse(source, mHandler);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.biojava.bio.seq.io.SequenceFormat#getDefaultFormat()
	 */
	@Override
	public String getDefaultFormat() {
		return SEQXML_FORMAT;
	}

	private PrintWriter pw;
	private XMLWriter xmlWriter;

	private XMLWriter getXMLWriter() {
		if (xmlWriter == null) {
			// make an XML writer
			pw = new PrintWriter(this.getPrintStream());
			xmlWriter = new PrettyXMLWriter(pw);
		}
		return xmlWriter;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.biojavax.bio.seq.io.RichSequenceFormat#beginWriting()
	 */
	@Override
	public void beginWriting() throws IOException {
		XMLWriter xml = getXMLWriter();
		xml.printRaw("<?xml version=\"1.0\" encoding=\"UTF-8\" ?>");
		xml.openTag(SEQXML_TAG);
		xml.attribute(SEQXML_SEQXMLVERSION_ATTR, SEQXML_VERSION);
		xml.attribute(XSI_SCHEMA_LOCATION_ATTR, String.format(SEQXML_SCHEMA,SEQXML_VERSION));
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.biojavax.bio.seq.io.RichSequenceFormat#finishWriting()
	 */
	@Override
	public void finishWriting() throws IOException {
		XMLWriter xml = getXMLWriter();
		xml.closeTag(SEQXML_TAG);
		pw.flush();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.biojava.bio.seq.io.SequenceFormat#writeSequence(org.biojava.bio.seq
	 * .Sequence, java.io.PrintStream)
	 */
	@Override
	public void writeSequence(Sequence seq, PrintStream os) throws IOException {
		if (this.getPrintStream() == null)
			this.setPrintStream(this.getPrintStream());
		this.writeSequence(seq, RichObjectFactory.getDefaultNamespace());
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.biojava.bio.seq.io.SequenceFormat#writeSequence(org.biojava.bio.seq
	 * .Sequence, java.lang.String, java.io.PrintStream)
	 */
	@Override
	public void writeSequence(Sequence seq, String format, PrintStream os)
			throws IOException {
		if (this.getPrintStream() == null)
			this.setPrintStream(this.getPrintStream());
		if (!format.equals(this.getDefaultFormat()))
			throw new IllegalArgumentException("Unknown format: " + format);
		this.writeSequence(seq, RichObjectFactory.getDefaultNamespace());
	}

	/**
	 * {@inheritDoc} Namespace is ignored as SeqXML has no concept of it.
	 */
	@Override
	public void writeSequence(Sequence seq, Namespace ns) throws IOException {
		RichSequence rs;
		try {
			if (seq instanceof RichSequence)
				rs = (RichSequence) seq;
			else
				rs = RichSequence.Tools.enrich(seq);
		} catch (ChangeVetoException e) {
			IOException e2 = new IOException("Unable to enrich sequence");
			e2.initCause(e);
			throw e2;
		}

		Set notes = rs.getNoteSet();

		XMLWriter xml = getXMLWriter();

		//Entry
		xml.openTag(ENTRY_TAG);
		xml.attribute(ENTRY_ID_ATTR, rs.getAccession());

		if(rs.getAnnotation().containsProperty(Terms.getSourceNameTerm())){
			
			String source = (String) rs.getAnnotation().getProperty(Terms.getSourceNameTerm());
			xml.attribute(ENTRY_SOURCE_ATTR,source);
		}
		
		//Species
		NCBITaxon tax = rs.getTaxon();
		if (tax != null) {
			xml.openTag(SPECIES_TAG);
			xml.attribute(SPECIES_NCBITAXID_ATTR, Integer.toString(tax.getNCBITaxID()));
			String name = "";
			for (Iterator i = tax.getNameClasses().iterator(); i.hasNext();) {
				String nameclass = (String) i.next();
				if (nameclass.equalsIgnoreCase(NCBITaxon.SCIENTIFIC))
					name = NCBITaxon.SCIENTIFIC;
				else if (nameclass.equalsIgnoreCase(SPECIES_NAMECLASS))
					name = SPECIES_NAMECLASS;
			}
			xml.attribute(SPECIES_NAME_ATTR, (String) tax.getNames(name).toArray()[0]);
			xml.closeTag(SPECIES_TAG);
		}

		//Description
		String description = rs.getDescription();
		if (description != null) {
			xml.openTag(DESCRIPTION_TAG);
			xml.print(description);
			xml.closeTag(DESCRIPTION_TAG);
		}

		//Sequence
		String alphabetName = rs.getAlphabet().getName();
		String seqTag = AASEQ_TAG;
		if (alphabetName.contains("DNA"))
			seqTag = DNASEQ_TAG;
		else if (alphabetName.contains("RNA"))
			seqTag = RNASEQ_TAG;
		xml.openTag(seqTag);
		String[] lines = StringTools.wordWrap(rs.seqString(), "\\s+", this
				.getLineWidth());
		for (int i = 0; i < lines.length; i++)
			xml.println(lines[i]);
		xml.closeTag(seqTag);

		//Database cross references
		if (!rs.getRankedCrossRefs().isEmpty()) {
			for (Iterator iterator = rs.getRankedCrossRefs().iterator(); iterator.hasNext();) {
				RankedCrossRef rankedCrossRef = (RankedCrossRef) iterator
						.next();
				CrossRef crossRef = rankedCrossRef.getCrossRef();
				String molType = "";
				for (Iterator j = crossRef.getRichAnnotation().getNoteSet()
						.iterator(); j.hasNext();) {
					Note n = (Note) j.next();
					if (n.getTerm().equals(Terms.getMolTypeTerm()))
						molType = n.getValue();
				}
				xml.openTag(DBREF_TAG);
				xml.attribute(DBREF_SEQTYPE_ATTR, molType);
				if (crossRef.getDbname() != null)
					xml.attribute(DBREF_SOURCE_ATTR, crossRef.getDbname());
				xml.attribute(DBREF_ID_ATTR, crossRef.getAccession());
				xml.closeTag(DBREF_TAG);
			}
		}

		if (!rs.getFeatureSet().isEmpty()) {
			for (Iterator iterator = rs.getFeatureSet().iterator(); iterator.hasNext();) {
				RichFeature feature = (RichFeature) iterator.next();
				String value = null;
				for (Iterator j = feature.getNoteSet().iterator(); j.hasNext();) {
					Note n = (Note) j.next();
					if (n.getTerm().equals(Terms.getFeatureValueTerm()))
						value = n.getValue();
				}
				xml.openTag(PROPERTY_TAG);
				xml.attribute(PROPERTY_NAME_ATTR, feature.getTypeTerm()
						.getName());
				if (value != null)
					xml.attribute(PROPERTY_VALUE_ATTR, value);
				xml.closeTag(PROPERTY_TAG);
			}
		}

		xml.closeTag(ENTRY_TAG);

	}

	private class SeqXMLHeaderHandler extends DefaultHandler {

		private String[] headerData;

		public SeqXMLHeaderHandler(String[] headerData) {
			this.headerData = headerData;
		}

		@Override
		public void startElement(String uri, String localName, String qName,
				Attributes attributes) throws SAXException {
			if (qName.equals(SEQXML_TAG)) {
				for (int i = 0; i < attributes.getLength(); i++) {
					String name = attributes.getQName(i);
					String val = attributes.getValue(i);
					if (name.equals(SEQXML_SOURCE_ATTR)) {
						headerData[0] = val;
					}
					if (name.equals(SEQXML_SOURCEVERSION_ATTR)) {
						headerData[1] = val;
					}
					if (name.equals(SEQXML_SEQXMLVERSION_ATTR)) {
						headerData[2] = val;
					}
				}
			}
		}

	}

	// SAX event handler for parsing SeqXML
	private class SeqXMLHandler extends DefaultHandler {
		
		private SymbolTokenization symParser;
		private RichSeqIOListener rlistener;
		private Namespace ns;
		private StringBuffer m_currentString;
		private NCBITaxon tax;
		private RichFeature.Template templ;
		private String[] headerData;

		private SeqXMLHandler(RichSequenceFormat parent,
				SymbolTokenization symParser, RichSeqIOListener rlistener,
				Namespace ns, String[] headerData) {
			this.symParser = symParser;
			this.rlistener = rlistener;
			this.ns = ns;
			this.m_currentString = new StringBuffer();
			this.headerData = headerData;
		}

		@Override
		public void startElement(String uri, String localName, String qName,
				Attributes attributes) throws SAXException {
			if (qName.equals(ENTRY_TAG)) {
				try {
					rlistener.startSequence();

					String sourceName = "";
					
					for (int i = 0; i < attributes.getLength(); i++) {
						String name = attributes.getQName(i);
						String val = attributes.getValue(i);
						if (name.equals(ENTRY_ID_ATTR)) {
							rlistener.setAccession(val);
							rlistener.setName(val);
						}
						else if (name.equals(ENTRY_SOURCE_ATTR)) {
							sourceName += val;
						}
					}
					if (ns == null)
						ns = RichObjectFactory.getDefaultNamespace();
					rlistener.setNamespace(ns);
					
					if (sourceName.isEmpty() && headerData[0] != null)
						sourceName += headerData[0];
					rlistener.addSequenceProperty(Terms.getSourceNameTerm(),
							sourceName);
					
					String version = headerData[1];
					if (version != null)
					{
						//BioJava expects the sequence version to be of type double
						try{
							
							Double versionDouble = Double.valueOf(version);
							rlistener.setSeqVersion(version);
						}
						catch(NumberFormatException e){}
					}
				} catch (ParseException e) {
					throw new SAXException(e);
				}
			} else if (qName.equals(SPECIES_TAG)) {
				Integer taxID = null;
				String taxName = null;
				for (int i = 0; i < attributes.getLength(); i++) {
					String name = attributes.getQName(i);
					String val = attributes.getValue(i);
					if (name.equals(SPECIES_NCBITAXID_ATTR))
						taxID = Integer.valueOf(val);
					if (name.equals(SPECIES_NAME_ATTR))
						taxName = val;
				}
				tax = (NCBITaxon) RichObjectFactory.getObject(
						SimpleNCBITaxon.class, new Object[] { taxID });
				tax.addName(SPECIES_NAMECLASS, taxName);
				try {
					rlistener.setTaxon(tax);
				} catch (ParseException e) {
					throw new SAXException(e);
				}

			} else if (qName.equals(DBREF_TAG)) {
				String source = "";
				String id = null;
				String type = null;
				for (int i = 0; i < attributes.getLength(); i++) {
					String name = attributes.getQName(i);
					String val = attributes.getValue(i);
					if (name.equals(DBREF_SEQTYPE_ATTR))
						type = val;
					else if (name.equals(DBREF_SOURCE_ATTR))
						source = val;
					else if (name.equals(DBREF_ID_ATTR))
						id = val;
				}
				CrossRef dbx = (CrossRef) RichObjectFactory.getObject(
						SimpleCrossRef.class, new Object[] { source, id,
								new Integer(0) });
				Note note = new SimpleNote(Terms.getMolTypeTerm(), type, 1);
				try {
					dbx.getRichAnnotation().addNote(note);
				} catch (ChangeVetoException ce) {
					SAXException pe = new SAXException(
							"Could not annotate identifier terms");
					pe.initCause(ce);
					throw pe;
				}
				RankedCrossRef rdbx = new SimpleRankedCrossRef(dbx, 0);
				try {
					rlistener.setRankedCrossRef(rdbx);
				} catch (ParseException e) {
					throw new SAXException(e);
				}

			} else if (qName.equals(PROPERTY_TAG)) {
				templ = new RichFeature.Template();
				templ.annotation = new SimpleRichAnnotation();
				templ.sourceTerm = Terms.getSeqXMLTerm();
				templ.featureRelationshipSet = new TreeSet();
				templ.rankedCrossRefs = new TreeSet();
				templ.location = new SimpleRichLocation(new SimplePosition(0),
						0);
				for (int i = 0; i < attributes.getLength(); i++) {
					String name = attributes.getQName(i);
					String val = attributes.getValue(i);
					if (name.equals(PROPERTY_NAME_ATTR)) {
						templ.typeTerm = RichObjectFactory.getDefaultOntology()
								.getOrCreateTerm(val);
					} else if (name.equals(PROPERTY_VALUE_ATTR)) {
						Note note = new SimpleNote(Terms.getFeatureValueTerm(),
								val, 1);
						((RichAnnotation) templ.annotation).addNote(note);
					}
				}
				try {
					// start the feature from the template we built
					rlistener.startFeature(templ);
					// end the feature
					rlistener.endFeature();
				} catch (ParseException e) {
					throw new SAXException(e);
				}
			}

		}

		@Override
		public void endElement(String uri, String localName, String qName)
				throws SAXException {
			String val = this.m_currentString.toString().trim();
			try {
				if (qName.equals(DESCRIPTION_TAG)) {
					rlistener.setDescription(val);
				} else if (qName.equals(RNASEQ_TAG) || qName.equals(DNASEQ_TAG)
						|| qName.equals(AASEQ_TAG)) {
					try {
						readSequence(val);
					} catch (Exception e) {
						throw new ParseException(e);
					}
				} else if (qName.equals(ENTRY_TAG)) {
					rlistener.endSequence();
				}
			} catch (ParseException e) {
				throw new SAXException(e);
			}

			// drop old string
			this.m_currentString.setLength(0);

		}

		private void readSequence(String val) throws Exception {
			SymbolList sl = new SimpleSymbolList(symParser, val.replaceAll(
					"\\s+", "").replaceAll("[\\.|~]", "-"));
			rlistener.addSymbols(symParser.getAlphabet(), (Symbol[]) (sl
					.toList().toArray(new Symbol[0])), 0, sl.length());
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see org.xml.sax.helpers.DefaultHandler#characters(char[], int, int)
		 */
		@Override
		public void characters(char[] ch, int start, int length)
				throws SAXException {
			this.m_currentString.append(ch, start, length);
		}

	}

}
